-- $Id$
Broker_Friends = LibStub("AceAddon-3.0"):NewAddon("Broker_Friends", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")

-- Grab used libraries.
local broker = LibStub("LibDataBroker-1.1")
local qtip = LibStub("LibQTip-1.0")

-- DB defaults.
local dbDefaults = {}

-- Tooltip holder.
local tooltip

-- Realm & faction data used for comparison. Filled OnEnable.
local playerRealm, playerFaction

-- Upvalues
local GetNumFriends, GetFriendInfo = GetNumFriends, GetFriendInfo
local BNFeaturesEnabledAndConnected, BNGetNumFriends, BNGetFriendInfo, BNGetNumFriendToons, BNGetFriendToonInfo =
	BNFeaturesEnabledAndConnected, BNGetNumFriends, BNGetFriendInfo, BNGetNumFriendToons, BNGetFriendToonInfo

-- Localized game client list. You would think there'd be a list of this somewhere...
local localizedGameClientList = {}
localizedGameClientList["UNKNOWN"] = UNKNOWN
localizedGameClientList[BNET_CLIENT_WOW] = "World of Warcraft"
localizedGameClientList[BNET_CLIENT_SC2] = "Starcraft 2"

-- Blizzard returns localized class names in the friends API. This is a huge pain, so we'll just make a list
-- of localized class names to class color codes here once for easy lookup.
local localizedClassColorTable = {}
for key, localized in pairs(LOCALIZED_CLASS_NAMES_MALE) do
	local color = RAID_CLASS_COLORS[key]
	localizedClassColorTable[localized] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end
for key, localized in pairs(LOCALIZED_CLASS_NAMES_FEMALE) do
	local color = RAID_CLASS_COLORS[key]
	localizedClassColorTable[localized] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end

function Broker_Friends:OnInitialize()
	-- Maybe in the future.
	--self.db = LibStub("AceDB-3.0"):New("Broker_FriendsDB", dbDefaults, "Default")
end

function Broker_Friends:OnEnable()
	self:RegisterEvent("FRIENDLIST_UPDATE", "Update")
	
	-- Make sure we have data and keep it somewhat up-to-date. The game is pretty decent with keeping it
	-- updated itself, so once a minute is more than enough.
	ShowFriends()
	self:ScheduleRepeatingTimer(ShowFriends, 60)
	-- Also, we set a one-time timer for 3 seconds because BN info takes a while to get through and may or
	-- may not decide to fire FRIENDLIST_UPDATE.
	self:ScheduleTimer(ShowFriends, 3)
	
	-- Initialize realm/faction data, it should be ok now.
	playerRealm = GetRealmName()
	playerFaction = UnitFactionGroup("player")
end

function Broker_Friends:OnDisable()
end

-- Update the text display and possibly the tooltip.
function Broker_Friends:Update()
	-- Basic friend count
	local normalFriends, onlineNormalFriends = GetNumFriends()
	
	local BNfriends, onlineBNfriends = 0
	if BNFeaturesEnabledAndConnected() then
		BNfriends, onlineBNfriends = BNGetNumFriends()
	end
	
	local totalOnline = onlineNormalFriends + onlineBNfriends
	local totalAll = normalFriends + BNfriends
	Broker_FriendsLDB.text = ("%d/%d"):format(totalOnline, totalAll)
	
	-- If we have a tooltip, great! Update it too, then.
	if qtip:IsAcquired(self) then
		self:UpdateTooltip()
	end
end

local charNameClicked = function(_, charName)
	SetItemRef(("player:%s"):format(charName), ("|Hplayer:%s|h[%s]|h"):format(charName, charName), "LeftButton")
end

-- QTip only allows you to pass 1 argument so we do it in a table.
local bnetNameClicked = function(_, tbl)
	local presenceID, name = unpack(tbl)
	local crap = ("%s:%d"):format(name, presenceID)
	SetItemRef(("BNplayer:%s"):format(crap), ("|HBNplayer:%s|h[%s]|h"):format(crap, name), "LeftButton")
end

-- Update the tooltip.
function Broker_Friends:UpdateTooltip()
	tooltip:Clear()
	
	tooltip:AddHeader("Name", "Level", "Area", "Note")
	for i = 1, GetNumFriends() do
		local name, level, class, area, connected, status, note = GetFriendInfo(i)
		
		if connected then
			local displayedName = name
			
			if status then
				displayedName = ("%s%s"):format(status, displayedName)
			end
			
			if localizedClassColorTable[class] then
				displayedName = ("%s%s%s"):format(localizedClassColorTable[class], displayedName, "|r")
			end
			
			local line = tooltip:AddLine(displayedName, level, area, note)
			tooltip:SetCellScript(line, 1, "OnMouseDown", charNameClicked, name)
		end
	end
	
	if BNFeaturesEnabledAndConnected() then
		-- Empty line between normal and RealID friends.
		tooltip:AddLine(" ")
		
		for friend = 1, BNGetNumFriends() do
			local presenceID, givenName, surname, bnCharName, _, _, isOnline, _,
				isAFK, isDND, _, noteText, _, _ = BNGetFriendInfo(friend)
			local fullName = BATTLENET_NAME_FORMAT:format(givenName, surname)
			local playerStatus = ""
			if isAFK then
				playerStatus = ("<%s>"):format(AFK)
			elseif isDND then
				playerStatus = ("<%s>"):format(DND)
			end
			
			if isOnline then
				-- Is a friend playing WoW? If so, we add his WoW toon(s) - if not, we add his RealID/SC2 info below.
				-- If a friend is playing WoW and SC2, this below code should only display his WoW toon - that's intended.
				local inWow = false
				local lastClient = "UNKNOWN"
				local numToons = BNGetNumFriendToons(friend)
				
				-- TODO: Test what happens when multiple toons are active.
				for toon = 1, numToons do
					local _, charName, gameClient, charRealm, factionNum, _, charClass, _,
						charArea, charLevel, gameText = BNGetFriendToonInfo(friend, toon)
					
					lastClient = gameClient
					
					-- *If* he's playing WoW, we add a proper toon line here.
					if gameClient == BNET_CLIENT_WOW then
						inWow = true
						
						-- Ok so, the API gives us back a faction number which we can turn into a real faction string
						-- by looking it up in this table. I have no idea if it's localized or not, so we just check
						-- both to be sure.
						local actualFaction = PLAYER_FACTION_GROUP[factionNum]

						-- Blizzard provides an API for this. Guess what? It uses the toonid which we don't have.
						-- Even the actual friends frame does it this way. We shouldn't have to do this...
						local canCooperate = false
						local displayedCharName = charName
						if charRealm == playerRealm and actualFaction == playerFaction then
							canCooperate = true
						else
							displayedCharName = ("%s-%s*"):format(charName, charRealm)
						end
						
						-- Status stuff.
						displayedCharName = playerStatus .. displayedCharName
						-- Color crap.
						if localizedClassColorTable[charClass] then
							displayedCharName = ("%s%s|r"):format(localizedClassColorTable[charClass], displayedCharName)
						end
						
						-- Finally, we get to add the line.
						local line = tooltip:AddLine(displayedCharName, charLevel, charArea, fullName)
						
						-- Set up whisper on click if we can "cooperate" (same faction/realm).
						if canCooperate then
							tooltip:SetCellScript(line, 1, "OnMouseDown", charNameClicked, charName)
						end

						-- Also set up B.net whispers.
						tooltip:SetCellScript(line, 4, "OnMouseDown", bnetNameClicked, {presenceID, fullName})
					end
				end
				
				-- If this friend isn't playing WoW at all (SC2 only), we add a different line here.
				-- TODO: Test this.
				if not inWow then
					local displayedCharName = bnCharName
					displayedCharName = ("%s%s%s|r"):format(FRIENDS_OTHER_NAME_COLOR_CODE, playerStatus, displayedCharName)
					
					-- This should be future-proof
					local client = localizedGameClientList[lastClient] or localizedGameClientList.UNKNOWN
					tooltip:AddLine(bnCharName, "-", ("In-game: %s"):format(client), fullName)
					tooltip:SetCellScript(line, 4, "OnMouseDown", bnetNameClicked, {presenceID, fullName})
				end
			end -- end online check
		end -- end bnet iterator
	end -- end bnet conditional
end

function Broker_Friends:ShowTooltipAt(anchor)
	ShowFriends()
	tooltip = qtip:Acquire(Broker_Friends, 4, "LEFT", "CENTER", "CENTER", "RIGHT")
	tooltip:SetAutoHideDelay(0.2, anchor)
	tooltip:SmartAnchorTo(anchor)
	tooltip:Show()
	self:UpdateTooltip()
end

-- Create the main data source
Broker_FriendsLDB = broker:NewDataObject("Broker_Friends", {
	type = "data source",
	text = "",
	icon = "Interface\\Icons\\INV_MISC_ORB_04",
	tocname = "Broker_Friends",
	
	OnEnter = function(anchor)
		Broker_Friends:ShowTooltipAt(anchor)
	end,
	
	OnClick = function()
		ToggleFriendsFrame(1)
	end,
})
